// ===========================================================================
//	Isadora Demo Plugin			   2003 Mark F. Coniglio. All rights reserved.
// ===========================================================================
//
//	IMPORTANT: This source code ("the software") is supplied to you in
//	consideration of your agreement to the following terms. If you do not
//	agree to the terms, do not install, use, modify or redistribute the
//	software.
//
//	Mark Coniglio (dba TroikaTronix) grants you a personal, non exclusive
//	license to use, reproduce, modify this software with and to redistribute it,
//	with or without modifications, in source and/or binary form. Except as
//	expressly stated in this license, no other rights are granted, express
//	or implied, to you by TroikaTronix.
//
//	This software is provided on an "AS IS" basis. TroikaTronix makes no
//	warranties, express or implied, including without limitation the implied
//	warranties of non-infringement, merchantability, and fitness for a 
//	particular purpurse, regarding this software or its use and operation
//	alone or in combination with your products.
//
//	In no event shall TroikaTronix be liable for any special, indirect, incidental,
//	or consequential damages arising in any way out of the use, reproduction,
//	modification and/or distribution of this software.
//
// ===========================================================================
//
// CUSTOMIZING THIS SOURCE CODE
// To customize this file, search for the text ###. All of the places where
// you will need to customize the file are marked with this pattern of 
// characters.
//
// ABOUT IMAGE BUFFER MAPS:
//
// The ImageBufferMap structure, and its accompanying functions,
// exists as a convenience to those writing video processing plugins.
//
// Basically, an image buffer contains an arbitrary number of input and
// output buffers (in the form of ImageBuffers). The ImageBufferMap code
// will automatically create intermediary buffers if needed, so that the
// size and depth of the source image buffers sent to your callback are
// the same for all buffers.
// 
// Typically, the ImageBufferMap is created in your CreateActor function,
// and dispose in the DiposeActor function.

// ---------------------------------------------------------------------------------
// INCLUDES
// ---------------------------------------------------------------------------------

#include "IsadoraTypes.h"
#include "IsadoraCallbacks.h"
#include "ImageBufferUtil.h"
#include "PluginDrawUtil.h"

// STANDARD INCLUDES
#include <string.h>
#include <stdio.h>

//PortAudio and sndfile includes
#include "portaudio.h"
#include <sndfile.h>

//Own includes
#include "alloc.h"
#include "lecture.h"

// ---------------------------------------------------------------------------------
// MacOS Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_MAC
#define EXPORT_
#endif

// ---------------------------------------------------------------------------------
// Win32  Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_WIN32

	#include <windows.h>
	
	#define EXPORT_ __declspec(dllexport)
	
	#ifdef __cplusplus
	extern "C" {
	#endif

	BOOL WINAPI DllMain ( HINSTANCE hInst, DWORD wDataSeg, LPVOID lpvReserved );

	#ifdef __cplusplus
	}
	#endif

	BOOL WINAPI DllMain (
		HINSTANCE	/* hInst */,
		DWORD		wDataSeg,
		LPVOID		/* lpvReserved */)
	{
	switch(wDataSeg) {
	
	case DLL_PROCESS_ATTACH:
		return 1;
		break;
	case DLL_PROCESS_DETACH:
		break;
		
	default:
		return 1;
		break;
	}
	return 0;
	}

#endif

// ---------------------------------------------------------------------------------
//	Exported Function Definitions
// ---------------------------------------------------------------------------------

#ifdef __cplusplus
extern "C" {
#endif

EXPORT_ void GetActorInfo(void* inParam, ActorInfo* outActorParams);

#ifdef __cplusplus
}
#endif


// ---------------------------------------------------------------------------------
//	FORWARD DECLARTIONS
// ---------------------------------------------------------------------------------
static void
ReceiveMessage(
	IsadoraParameters*	ip,
	MessageMask			inMessageMask,
	PortIndex			inPortIndex1,
	const MsgData*		inData,
	UInt32				inLen,
	long				inRefCon);

// ---------------------------------------------------------------------------------
// GLOBAL VARIABLES
// ---------------------------------------------------------------------------------
// ### Declare global variables, common to all instantiations of this plugin here

static int gSampleRate=44100;//#*# Sample Rate
//#*# J'assume que tous les fichiers qu'on reproduira ont  peu prs le mme 
// sample rate, si c'est pas prcisement 44100 on entendra le son un peu plus lent ou un peu plus 
// rapide que l'original
static PaDeviceIndex gOutDevice=3; //#*# Output device index found by inspecting answer from pa_devs.exe
static PaTime gSugLat=0.0100; //#*# Suggested latency

// Example: static int gMyGlobalVariable = 5;

static int gChCount=8 ;// #*# 
static int gChWav=2;// #*# 
static int gFrPerBuf=7350 ;// #*# 
#define FR_PER_BUF 7350 // #*# rajout pour pouvoir definir le buffer dbuf comme un tableau de taille constante
//#*# cette valeur doit tre legerement plus grande que
//#*# le Sample Rate divise par le fps defini dans les prefs d'isadora
#define CH_COUNT 8 // #*# rajout pour pouvoir definir les buffers buf et dbuf comme des tableaux de taille constante
#define CH_WAV 2// #*# rajout pour pouvoir definir les buffers buf et dbuf comme des tableaux de taille constante
//#*# J'assume que chacun des fichiers qui sera reproduit est un fichier wav stereo donc avec deux canaux

// ---------------------------------------------------------------------------------
// PluginInfo struct
// ---------------------------------------------------------------------------------
// ### This structure neeeds to contain all variables used by your plugin. Memory for
// this struct is allocated during the CreateActor function, and disposed during
// the DisposeActor function, and is private to each copy of the plugin.
//
// If your plugin needs global data, declare them as static variables within this
// file. Any static variable will be global to all instantiations of the plugin.

typedef struct {

	ActorInfo*				mActorInfoPtr;		// our ActorInfo Pointer - set during create actor function
	MessageReceiverRef		mMessageReceiver;	// pointer to our message receiver reference
	Boolean					mNeedsDraw;			// set to true when the video output needs to be drawn
	

	UInt32					mSoundIndex[4];			// #*# The Number of the Wav file to reproduce in Media Window
	Boolean					mLoopEnable[4];	//#*# Indicates whether the sound will be played once or many times
	float					mVolume[4];		//#*# Indicates the volume
	Value					mPosition[4];	//#*# Indicates the position inside the sound

		//#*# Private members
	SInt32					mLeft[4];		// #*# Left Channel Number , ranges from 0 to gChCount-1=7
	SInt32					mRight[4];		// #*# Right Channel Number, ranges from 0 to gChCount-1=7
	SF_INFO sfi[4];		//#*# garde l'information du fichier
	SNDFILE *sf[4];			//#*# c'est le fichier intermdiaire de lecture
//	SNDFILE *sfinit[4];		//#*# c'est le fichier initial de lecture
	sf_count_t nread[4];		//#*# compte le nombre de frames significatifs dans buf
	float buf[4][FR_PER_BUF*CH_WAV];//#*# ce buffer sera utilis pour la lecture audio du fichier
	float dbuf[FR_PER_BUF*CH_COUNT];//#*# ce buffer sera utilis pour transporter l'audio remapp
	PaStream *stream;		//#*# c'est le flux audio de portaudio

	char fileName[4][256];//#*# Le nom du fichier reproduit 
	bool isFileOpen[4];//#*# signale si le fichier audio est ouvert
	bool isPlaying[4];//#*# signale si la lecture du fichier audio est en cours,
    bool isDeviceRecognized;//#*# signale si les MOTUs ont t reconnues

	//#*# champs rajouts pour le control, une fois le plugin en route il faudra les enlever
	PaError err;//#*# nous informera si les fonctions PortAudio ont bien march
	const PaDeviceInfo *deviceInfo;//#*#information sur la device utilise
	
} PluginInfo;

// A handy macro for casting the mActorDataPtr to PluginInfo*
#if __cplusplus
#define	GetPluginInfo_(actorDataPtr)		static_cast<PluginInfo*>((actorDataPtr)->mActorDataPtr);
#else
#define	GetPluginInfo_(actorDataPtr)		(PluginInfo*)((actorDataPtr)->mActorDataPtr);
#endif

// ---------------------------------------------------------------------------------
//	Constants
// ---------------------------------------------------------------------------------
//	Defines various constants used throughout the plugin.

// ### GROUP ID
// Define the group under which this plugin will be displayed in the Isadora interface.
// These are defined under "Actor Types" in IsadoraTypes.h

static const OSType	kActorClass 	= kGroupSound;

// ### PLUGIN IN
// Define the plugin's unique four character identifier. Contact TroikaTronix to
// obtain a unique four character code if you want to ensure that someone else
// has not developed a plugin with the same code. Note that TroikaTronix reserves
// all plugin codes that begin with an underline, an at-sign, and a pound sign
// (e.g., '_', '@', and '#'.)

static const OSType	kActorID		= FOUR_CHAR_CODE('AuC1');

// ### ACTOR NAME
// The name of the actor. This is the name that will be shown in the User Interface.

static const char* kActorName		= "MOTU #1 (Com)";

// ### PROPERTY DEFINITION STRING
// The property string. This string determines the inputs and outputs for your plugin.
// See the IsadoraCallbacks.h under the heading "PROPERTY DEFINITION STRING" for the
// meaning ofthese codes. (The IsadoraCallbacks.h header can be seen by opening up
// the IzzySDK Framework while in the Files view.)
//
// IMPORTANT: You cannot use spaces in the property name. Instead, use underscores (_)
// where you want to have a space.
//
// Note that each line ends with a carriage return (\r), and that only the last line of
// the bunch ends with a semicolon. This means that what you see below is one long
// null-terminated c-string, with the individual lines separated by carriage returns.

static const char* sPropertyDefinitionString =

// INPUT PROPERTY DEFINITIONS
//	TYPE 	PROPERTY NAME	ID		DATATYPE	DISPLAY FMT			MIN		MAX		INIT VALUE
"INPROP sound1			sin1		int			number				0		*		0\r"
//"INPROP left1			lef1		int			number				1		8		1\r"//#*#gChCount=8
//"INPROP right1			rig1		int			number				1		8		2\r"//#*#gChCount=8
"INPROP loop1			loo1		bool		onoff				0		1		0\r"
"INPROP volume1			vol1		float		number				0		200		100\r"
"INPROP start1			sta1		bool		trig				0		1		0\r"
"INPROP sound2			sin2		int			number				0		*		0\r"
//"INPROP left2			lef2		int			number				1		8		3\r"//#*#gChCount=8
//"INPROP right2			rig2		int			number				1		8		4\r"//#*#gChCount=8
"INPROP loop2			loo2		bool		onoff				0		1		0\r"
"INPROP volume2			vol2		float		number				0		200		100\r"
"INPROP start2			sta2		bool		trig				0		1		0\r"
"INPROP sound3			sin3		int			number				0		*		0\r"
//"INPROP left3			lef3		int			number				1		8		5\r"//#*#gChCount=8
//"INPROP right3			rig3		int			number				1		8		6\r"//#*#gChCount=8
"INPROP loop3			loo3		bool		onoff				0		1		0\r"
"INPROP volume3			vol3		float		number				0		200		100\r"
"INPROP start3			sta3		bool		trig				0		1		0\r"
"INPROP sound4			sin4		int			number				0		*		0\r"
//"INPROP left4			lef4		int			number				1		8		7\r"//#*#gChCount=8
//"INPROP right4			rig4		int			number				1		8		8\r"//#*#gChCount=8
"INPROP loop4			loo4		bool		onoff				0		1		0\r"
"INPROP volume4			vol4		float		number				0		200		100\r"
"INPROP start4			sta4		bool		trig				0		1		0\r"

// OUTPUT PROPERTY DEFINITIONS
//	TYPE 	 PROPERTY NAME	ID		DATATYPE	DISPLAY FMT			MIN		MAX		INIT VALUE
"OUTPROP position1		pos1		float		number			0		100		0\r"
"OUTPROP end1			end1		bool		trig			0		1		0\r"
"OUTPROP position2		pos2		float		number			0		100		0\r"
"OUTPROP end2			end2		bool		trig			0		1		0\r"
"OUTPROP position3		pos3		float		number			0		100		0\r"
"OUTPROP end3			end3		bool		trig			0		1		0\r"
"OUTPROP position4		pos4		float		number			0		100		0\r"
"OUTPROP end4			end4		bool		trig			0		1		0\r"
;
// ### Property Index Constants
// Properties are referenced by a one-based index. The first input property will
// be 1, the second 2, etc. Similarly, the first output property starts at 1.
// You whould have one constant for each input and output property defined in the 
// property definition string.

enum
{
	kInputSoundNum1 = 1,
//	kInputLeft1,
//	kInputRight1,
	kInputLoopEnable1,
	kInputVolume1,
	kInputStart1,
	kInputSoundNum2,
//	kInputLeft2,
//	kInputRight2,
	kInputLoopEnable2,
	kInputVolume2,
	kInputStart2,
	kInputSoundNum3,
//	kInputLeft3,
//	kInputRight3,
	kInputLoopEnable3,
	kInputVolume3,
	kInputStart3,
	kInputSoundNum4,
//	kInputLeft4,
//	kInputRight4,
	kInputLoopEnable4,
	kInputVolume4,
	kInputStart4,

	kOutputPosition1 = 1,
    kOutputEnd1,
	kOutputPosition2,
	kOutputEnd2,
	kOutputPosition3,
	kOutputEnd3,
	kOutputPosition4,
	kOutputEnd4,
};


// ---------------------
//	Help String
// ---------------------
// ### Help Strings
//
// The first help string is for the actor in general. This followed by help strings
// for all of the inputs, and then by the help strings for all of the outputs. These
// should be given in the order that they are defined in the Property Definition
// String above.
//
// In all, the total number of help strings should be (num inputs + num outputs + 1)
//
// Note that each string is followed by a comma -- it is a common mistake to forget the
// comma which results in the two strings being concatenated into one.

const char* sHelpStrings[] =
{
	"Plays 4 given wav files to chosen channels.",

	"The number of the first sound, as shown in the Media window, that will be played by this actor.",

//	"The number of the channel where the left part of the first stereo sound will be played.",

//	"The number of the channel where the right part of the first stereo sound will be played.",
		
	"If true, the first sound is played again and again. Otherwise it is played only once.",

	"The volume of the  first sound. Zero is mute, 100 is normal volume. Warning: overpassing 100 could cause distorsion of the sound",

	"The trigger to start playing the  first sound.",

	"The number of the second sound, as shown in the Media window, that will be played by this actor.",

//	"The number of the channel where the left part of the second stereo sound will be played.",

//	"The number of the channel where the right part of the second stereo sound will be played.",
		
	"If true, the second sound is played again and again. Otherwise it is played only once.",

	"The volume of the second sound. Zero is mute, 100 is normal volume. Warning: overpassing 100 could cause distorsion of the sound",

	"The trigger to start playing the  second sound.",

	"The number of the third sound, as shown in the Media window, that will be played by this actor.",

//	"The number of the channel where the left part of the third stereo sound will be played.",

//	"The number of the channel where the right part of the third stereo sound will be played.",	

	"If true, the third sound is played again and again. Otherwise it is played only once.",

	"The volume of the third sound. Zero is mute, 100 is normal volume. Warning: overpassing 100 could cause distorsion of the sound",

	"The trigger to start playing the third sound.",

	"The number of the fourth sound, as shown in the Media window, that will be played by this actor.",

//	"The number of the channel where the left part of the fourth stereo sound will be played.",

//	"The number of the channel where the right part of the fourth stereo sound will be played.",
	
	"If true, the fourth sound is played again and again. Otherwise it is played only once.",

	"The volume of the fourth sound. Zero is mute, 100 is normal volume. Warning: overpassing 100 could cause distorsion of the sound",

	"The trigger to start playing the  fourth sound.",

	"Indicates how much of the first sound has already been played.",

	"A trigger is send when first sound arrive to the end.",
	
	"Indicates how much of the second sound has already been played.",

	"A trigger is send when second sound arrive to the end.",
	
	"Indicates how much of the third sound has already been played.",

	"A trigger is send when third sound arrive to the end.",

	"Indicates how much of the fourth sound has already been played.",

	"A trigger is send when fourth sound arrive to the end.",
};

// ---------------------------------------------------------------------------------
//		 CreateActor
// ---------------------------------------------------------------------------------
// Called once, prior to the first activation of an actor in its Scene. The
// corresponding DisposeActor actor function will not be called until the file
// owning this actor is closed, or the actor is destroyed as a result of being
// cut or deleted.

static void
CreateActor(
	IsadoraParameters*	ip,	
	ActorInfo*			ioActorInfo)		// pointer to this actor's ActorInfo struct - unique to each instance of an actor
{
	// create the PluginInfo struct - initializing it to all zeroes
	PluginInfo* info = (PluginInfo*) IzzyMallocClear_(ip, sizeof(PluginInfo));
	PluginAssert_(ip, info != nil);
	
	ioActorInfo->mActorDataPtr = info;
	info->mActorInfoPtr = ioActorInfo;

	// ### allocation of private member variables
//	info->mSoundIndex=(UInt32*)malloc(4*sizeof(UInt32));
//	info->mLeft=(SInt32*)malloc(4*sizeof(UInt32));
//	info->mRight=(SInt32*)malloc(4*sizeof(UInt32));

//	info->sfi=(SF_INFO*)malloc(4 * sizeof(SF_INFO));	
	for(int i=0;i<4;i++){
		info->sf[i]=(SNDFILE_tag*)malloc(sizeof(SNDFILE_tag*));	
//		info->sfinit[i]=(SNDFILE_tag*)malloc(sizeof(SNDFILE_tag*));	
	}//#*# !!!Comment momentanement car faisait planter la fermeture d'ISadora

//	info->nread=(sf_count_t*)malloc(4 * sizeof(sf_count_t));	
//	info->buf=Make2DFloatArray(4,gFrPerBuf*gChWav);//#*# ce buffer sera utilis pour la lecture audio du fichier
//	info->dbuf=(float*)malloc(gFrPerBuf*gChCount*sizeof(float));//#*# ce buffer sera utilis pour transporter l'audio remapp

//	info->fileName=Make2DCharArray(4,256);
//	info->isFileOpen=(bool*)malloc(4*sizeof(bool));//#*# signale si le fichier audio est ouvert
//	info->isPlaying=(bool*)malloc(4*sizeof(bool));//#*# signale si la lecture du fichier audio est en cours,
//#*# !!! si on revient  des tableaux dynamiques il faudra decommenter les lignes ci-dessus
	//#*# !!! info->stream n'est pas allou

	// ### initialization of private member variables
	for(int i=0;i<4;i++){
		info->mSoundIndex[i]=(UInt32)0;//#*#Valeur par defaut: aucun fichier
		info->mLeft[i]=2*i;//#*#Valeur par defaut: son gauche par canal pair
		info->mRight[i]=2*i+1;//#*#Valeur par defaut: son droit par canal impair
		info->mPosition[i].type=kFloat;
		info->mPosition[i].u.fvalue=0;
		info->sfi[i].frames=0;
		info->sfi[i].channels=0;
		info->sfi[i].format=0;
		info->sfi[i].samplerate=0;
		info->sfi[i].sections=0;
		info->sfi[i].seekable=0;
		//#*# !!! info->sf n'est pas initialis, par contre il est bien allou ci-dessus
		info->nread[i]=0;
		//#*# Fills buffers with null sound
		for(int k=0;k<gFrPerBuf;k++){
			for(int ch=0;ch<gChWav;ch++){
				info->buf[i][k*gChWav+ch]=0;//#*# Fills the file's buffer with null sound
			}
			for(int ch=0;ch<gChCount;ch++){
				info->dbuf[k*gChCount+ch]=0;//#*# Fills the deinterleaved buffer with null sound
			}
		}
		//#*# !!! info->stream n'est pas initialis
		for(int k=0;k<256;k++){
			info->fileName[i][k]='\0';
		}
		//#*# initialise les boolens  0
		info->isPlaying[i]=false;
		info->isFileOpen[i]=false;

	}
	
	//#*# champs rajouts pour le control, une fois le plugin en route il faudra les enlever
	info->isDeviceRecognized=false;//#*# au debut la MOTU n'est pas encore reconnue
	info->err=paNoError;//#*# puisqu'ici on n'a pas encore commenc  utiliser portaudio, il n'y a pas d'erreur

}

// ---------------------------------------------------------------------------------
//		 DisposeActor
// ---------------------------------------------------------------------------------
// Called when the file owning this actor is closed, or when the actor is destroyed
// as a result of its being cut or deleted.
//
static void
DisposeActor(
	IsadoraParameters*	ip,
	ActorInfo*			ioActorInfo)		// pointer to this actor's ActorInfo struct - unique to each instance of an actor
{
	PluginInfo* info = GetPluginInfo_(ioActorInfo);
	PluginAssert_(ip, info != nil);
	
	//#*# Chaque ouverture de fichier avec sf_open doit tre suivie d'une fermeture avec sf_close
	for(int i=0;i<4;i++){
		if(info->isFileOpen[i]){
			sf_close(info->sf[i]);//#*#if the SNDFILE is not closed, it close it
			info->isFileOpen[i]=false;
		}
	}


	// ### destruction of private member variables
	
	// #*# destruction of buffers
	//free(info->mSoundIndex);
	//free(info->mLeft);
	//free(info->mRight);
	//free(info->sfi);
	free(info->sf);//#*# !!!Comment momentanement car faisait planter la fermeture d'ISadora
//	free(info->sfinit);//#*# !!!Comment momentanement car faisait planter la fermeture d'ISadora
	//free(info->nread);
	//Free2DFloatArray(info->buf,4);
	//free(info->dbuf);
//	Free2DCharArray(info->fileName,4); //#*# cette ligne a t change car elle causait des plantages
	//free(info->fileName);
	//free(info->isFileOpen);
	//free(info->isPlaying);
//#*# !!! si on revient  des tableaux dynamiques il faudra decommenter les lignes ci-dessus

	// destroy the PluginInfo struct allocated with IzzyMallocClear_ the CreateActor function
	PluginAssert_(ip, ioActorInfo->mActorDataPtr != nil);
	IzzyFree_(ip, ioActorInfo->mActorDataPtr);
}

// ---------------------------------------------------------------------------------
//		 ActivateActor
// ---------------------------------------------------------------------------------
//	Called when the scene that owns this actor is activated or deactivated. The
//	inActivate flag will be true when the scene is activated, false when deactivated.
//
static void
ActivateActor(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,		// pointer to this actor's ActorInfo struct - unique to each instance of an actor
	Boolean				inActivate)			// true when actor is becoming active, false otherwise.
{
	PluginInfo* info = GetPluginInfo_(inActorInfo);
	
	// ------------------------
	// ACTIVATE
	// ------------------------
	
	if (inActivate) {
	
		// Isadora passes various messages to plugins that request them.
		// These include Mouse Moved messages, Key Down/Key Up messages,
		// Video Frame Clock messages, etc. The complete list can be found
		// in the enumeration in MessageReceiverCommon.h
		
		// You ask Isadora for these messages by calling CreateMessageReceiver_
		// with a pointer to your function, and the message types you would
		// like to receive. (These are bitmapped flags, so you can combine as
		// many as you like: kWantKeyDown | kWantKeyDown for instance.)
		
		// Here we request that our ReceiveMessage function is called
		// whenever the Isadora New Video Frame message is sent,
		// which happens periodically, 30 times per second. We set the ref
		// con to our ActorInfo ptr so that we can access that information
		// from ReceiveMessage callback.
		
		MessageReceiveFunction* msgRcvFunc = ReceiveMessage;
		
		// #*# We want to receive messages
		{		
		// we should not already have a message receiver
		PluginAssert_(ip, info->mMessageReceiver == nil);
			
		// create a message receiver that will be notified of
		// video frame ticks
		info->mMessageReceiver = CreateMessageReceiver_(
			ip,
			msgRcvFunc,
			0,
			kWantVideoFrameTick,
			(long) inActorInfo);
		}
		
		// set the needs draw flag so that we will be drawn as soon
		// as possible
		info->mNeedsDraw = false; //#*# je change cette valeur  false car nous n'avons pas de sortie video

		//
		//#*# Start Portaudio
		info->err=Pa_Initialize();
		int b=-1; //#*# variable de control pour dire quand sortir de la boucle while qui cherche la motu
		int numDevices = Pa_GetDeviceCount();
		int j=-1;
		while(b<0){
			j++;
			if(j<numDevices){ //#*# tant qu'on parcours les devices 
				info->deviceInfo=Pa_GetDeviceInfo(j);
				if(info->deviceInfo->maxInputChannels == 0){
					if(info->deviceInfo->maxOutputChannels == 8){
						if(info->deviceInfo->name[27]=='1'){ //#*# a correspond  la MOTU #1
							if(info->deviceInfo->hostApi==Pa_HostApiTypeIdToHostApiIndex(paWASAPI)){
								b=0;
							}
						}
					}
				}
			}else{ 
				b=0;
			}
		}
		if(j==numDevices){
			info->isDeviceRecognized=false;
		}
		else{
			info->isDeviceRecognized=true;
			PaStreamParameters * outputParameters=(PaStreamParameters*)malloc(sizeof(PaStreamParameters));//#*# ces paramtres determineront comment portaudio gre la sortie:
			outputParameters->channelCount = gChCount;//#*#gChCount=8
			outputParameters->device=j;
			outputParameters->hostApiSpecificStreamInfo = NULL;//#*#
			outputParameters->sampleFormat = paFloat32;//#*#
			outputParameters->suggestedLatency = gSugLat;//#*#

			/*#*# Assume that the audio device can handle the input file's sample rate
			and number of channels */
			info->err=Pa_OpenStream(
				&info->stream, 
				NULL, /*#*# no input */              
				outputParameters,
				gSampleRate,
				paFramesPerBufferUnspecified,
				paNoFlag,
				NULL, /*#*# no callback, use blocking API */
				NULL);/*#*# no callback, so no callback userData */
			info->err=Pa_StartStream(info->stream);//#*# starts stream
		}
	
	// ------------------------
	// DEACTIVATE
	// ------------------------
	
	} else {
	
		// dispose our message receiver when we are deactivated.
		if (info->mMessageReceiver != nil) {
			DisposeMessageReceiver_(ip, info->mMessageReceiver);
			info->mMessageReceiver = nil;
			info->mNeedsDraw = false;//#*# je change cette valeur  false car nous n'avons pas de sortie video
			                         // a la base l'assignation ce faisait avec |= mais je ne sais pas ce que a veut dire
		}
		
		// ### dispose any data that you don't need when 
		// you are not active.

		//#*# Stop portaudio
		if(info->isDeviceRecognized){ //#*# On a ouvert un flux sonore si et seulement si la MOTU a t reconnue
			info->err=Pa_StopStream(info->stream);//#*#
			info->err=Pa_CloseStream(info->stream);//#*#
			for(int i=0;i<4;i++){ //#*# puisque la device a t reconnue, on a peut-tre ouvert des fichiers, on les ferme ici.
				if(info->isFileOpen[i]){
					sf_close(info->sf[i]);
					info->isFileOpen[i]=false;
				}
			}
		}
		info->err=Pa_Terminate();//#*#
	}
}

// ---------------------------------------------------------------------------------
//		 GetParameterString
// ---------------------------------------------------------------------------------
//	Returns the property definition string. Called when an instance of the actor
//	needs to be instantiated.

static const char*
GetParameterString(
	IsadoraParameters*	/* ip */,
	ActorInfo*			/* inActorInfo */)
{
	return sPropertyDefinitionString;
}

// ---------------------------------------------------------------------------------
//		 GetHelpString
// ---------------------------------------------------------------------------------
//	Returns the help string for a particular property. If you have a fixed number of
//	input and output properties, it is best to use the PropertyTypeAndIndexToHelpIndex_
//	function to determine the correct help string to return.

static void
GetHelpString(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyType		inPropertyType,			// kPropertyTypeInvalid when requesting help for the actor
												// or kInputProperty or kOutputProperty when requesting help for a specific property
	PropertyIndex		inPropertyIndex1,		// the one-based index of the property (when inPropertyType is not kPropertyTypeInvalid)
	char*				outParamaterString,		// receives the help string
	UInt32				inMaxCharacters)		// size of the outParamaterString buffer
{
	const char* helpstr = nil;
	
	// The PropertyTypeAndIndexToHelpIndex_ converts the inPropertyType and
	// inPropertyIndex1 parameters to determine the zero-based index into
	// your list of help strings.
	UInt32 index1 = PropertyTypeAndIndexToHelpIndex_(ip, inActorInfo, inPropertyType, inPropertyIndex1);
	
	// get the help string
	helpstr = sHelpStrings[index1];
	
	// copy it to the output string
	strncpy(outParamaterString, helpstr, inMaxCharacters);
}
	
// ---------------------------------------------------------------------------------
//		 HandlePropertyChangeValue	[INTERRUPT SAFE]
// ---------------------------------------------------------------------------------
//	### This function is called whenever one of the input values of an actor changes.
//	The one-based property index of the input is given by inPropertyIndex1.
//	The new value is given by inNewValue, the previous value by inOldValue.
//
static void
HandlePropertyChangeValue(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyIndex		inPropertyIndex1,			// the one-based index of the property than changed values
	ValuePtr			/* inOldValue */,			// the property's old value
	ValuePtr			inNewValue,					// the property's new value
	Boolean				/* inInitializing */)		// true if the value is being set when an actor is first initalized
{
	PluginInfo* info = GetPluginInfo_(inActorInfo);

	// ### When you add/change/remove properties, you will need to add cases
	// to this switch statement, to process the messages for your
	// input properties
	
	// The value comes to you encapsulated in a Value structure. See 
	// ValueCommon.h for details about the contents of this structure.
	
	switch (inPropertyIndex1) {
	case kInputSoundNum1:
		{//#*# Set new song number
			info->mSoundIndex[0] = (int)(inNewValue->u.ivalue);
			if(info->mPosition[0].u.fvalue!=0){//#*#on retrace la position  zro
				info->mPosition[0].u.fvalue=0; 
				SetOutputPropertyValue_(ip, inActorInfo, kOutputPosition1, &(info->mPosition[0]));
			}
			if(info->mSoundIndex[0]>0){//#*# !!! Ici il faudrait aussi ajouter comme maximum la 
				                       //quantit totale de fichiers audio dans la liste de la fenetre 
				                       //media d'ISadora. A l'heure qu'il est je ne sais pas comment recuperer
				                       //ce nombre, et passer un nombre plus grand que cleui l fait bouger 
				                       //la fonction DisposeActor, dans la ligne
				                       //Free2DCharArray(info->fileName,4) car filename a un format inattendu aprs
				                       //qu'on essaye d'y crir un nom de fichier qui n'existe pas.
				//#*# si on avait laisser un fichier ouvert  cet endroit, ne le ferme d'abord.
	
				info->fileName[0][255]='\0';
				ExecuteMediaCommand_(ip, inActorInfo, 0, kMediaGetName, kMediaTypeAudio, info->mSoundIndex[0], (long) info->fileName[0],sizeof(info->fileName[0]));
				//#*# Cette partie a t amelior par Mark par message priv sur le forum. Elle 
				//#*# permet de rcuperer la route du nom d'un fichier  partir de l'index
				//#*# On rcupre ici la route et le nom du fichier

				char fpBuf[1024];

				Boolean success = GetUTF8PathToCurrentDocument_(ip, inActorInfo->mThing, sizeof(fpBuf), fpBuf);
				if (success==TRUE && IsRelativePathName(info->fileName[0])) {
					//strncat(fpBuf, info->fileName[0], sizeof(fpBuf));
					sprintf(fpBuf, "%s%s",fpBuf, info->fileName[0]);
					//strncpy(info->fileName[0],fpBuf, sizeof(fpBuf));
					sprintf(info->fileName[0],"%s",fpBuf);
				}

				if((info->sf[0] = sf_open(info->fileName[0], SFM_READ, &info->sfi[0])) != NULL){
					info->isFileOpen[0]=true;
				}
				else{
					sf_close(info->sf[0]);
					info->isFileOpen[0]=false;
				}
			}else{
				info->isPlaying[0]=false;
			}
			if(!(info->isFileOpen[0])){//#*# si la reproduction d'aucun fichier n'arrive par la suite.
				info->isPlaying[0]=false;
				//#*# Nous nettoyons les canaux attribus  ce fichier, s'ils sont valides
				if((0<(info->mLeft[0]+1))&&(info->mLeft[0]<gChCount)){
					for(int k=0;k<gFrPerBuf;k++){
						info->dbuf[k*gChCount+info->mLeft[0]]=0;
					}
				}
				if((0<(info->mRight[0]+1))&&(info->mRight[0]<gChCount)){
					for(int k=0;k<gFrPerBuf;k++){
						info->dbuf[k*gChCount+info->mRight[0]]=0;
					}
				}
			}
		}
		break; 

/*	case kInputLeft1:
		{//#*# Clear the entries in the buffer for the old channel
			for(int k=0;k<gFrPerBuf;k++){
				info->dbuf[k*gChCount+info->mLeft[0]]=0;
			}
			//#*# Set new left channel number
			info->mLeft[0] = (inNewValue->u.ivalue-1);
		}
		break;
	
	case kInputRight1:
		{//#*# Clear the entries in the buffer for the old channel
			for(int k=0;k<gFrPerBuf;k++){
				info->dbuf[k*gChCount+info->mRight[0]]=0;
			}
			//#*# Set new right channel number
			info->mRight[0] = (inNewValue->u.ivalue-1);
		}
		break;*/
		
	case kInputLoopEnable1:
		{//#*# Set new boolean value
			info->mLoopEnable[0] = (inNewValue->u.ivalue);
		}
		break;

	case kInputVolume1:
		{//#*# Set new volume value
			info->mVolume[0] = (inNewValue->u.fvalue);
		}
		break;

	case kInputStart1:
		{	if(!(info->isFileOpen[0])){
				if((info->sf[0] = sf_open(info->fileName[0], SFM_READ, &info->sfi[0])) == NULL){
					sf_close(info->sf[0]);
					info->isPlaying[0]=false;
				}else{
					info->isFileOpen[0]=true;
					info->isPlaying[0]=true;
				}
			}else{
				info->isPlaying[0]=true;
			}
		}
		break;

	case kInputSoundNum2:
		{//#*# Set new song number
			info->mSoundIndex[1] = (int)(inNewValue->u.ivalue);
			if(info->mPosition[1].u.fvalue!=0){//#*#on retrace la position  zro
				info->mPosition[1].u.fvalue=0; 
				SetOutputPropertyValue_(ip, inActorInfo, kOutputPosition2, &(info->mPosition[1]));
			}
			if(info->mSoundIndex[1]>0){//#*# !!! Ici il faudrait aussi ajouter comme maximum la 
				                       //quantit totale de fichiers audio dans la liste de la fenetre 
				                       //media d'ISadora. A l'heure qu'il est je ne sais pas comment recuperer
				                       //ce nombre, et passer un nombre plus grand que cleui l fait bouger 
				                       //la fonction DisposeActor, dans la ligne
				                       //Free2DCharArray(info->fileName,4) car filename a un format inattendu aprs
				                       //qu'on essaye d'y crir un nom de fichier qui n'existe pas.
				//#*# si on avait laisser un fichier ouvert  cet endroit, ne le ferme d'abord.
				if(info->isFileOpen[1]){
					sf_close(info->sf[1]);
					info->isFileOpen[1]=false;
				}
				info->fileName[1][255]='\0';
				ExecuteMediaCommand_(ip, inActorInfo, 0, kMediaGetName, kMediaTypeAudio, info->mSoundIndex[1], (long) info->fileName[1], sizeof(info->fileName[1]));

				//#*# Cette partie a t amelior par Mark par message priv sur le forum. Elle 
				//#*# permet de rcuperer la route du nom d'un fichier  partir de l'index
				//#*# On rcupre ici la route et le nom du fichier

				char fpBuf[1024];
				Boolean success = GetUTF8PathToCurrentDocument_(ip, inActorInfo->mThing, sizeof(fpBuf), fpBuf);
				if (success==TRUE && IsRelativePathName(info->fileName[1])) {
					//strncat(fpBuf, info->fileName[1], sizeof(fpBuf));
					sprintf(fpBuf, "%s%s",fpBuf, info->fileName[1]);
					//strncpy(info->fileName[1],fpBuf, sizeof(fpBuf));
					sprintf(info->fileName[1],"%s",fpBuf);
				}
				if((info->sf[1] = sf_open(info->fileName[1], SFM_READ, &info->sfi[1])) != NULL){
					info->isFileOpen[1]=true;
				}
				else{
					sf_close(info->sf[1]);
					info->isFileOpen[1]=false;
				}
			}else{
				info->isPlaying[1]=false;
			}
			if(!(info->isFileOpen[1])){//#*# si la reproduction d'aucun fichier n'arrive par la suite.
				info->isPlaying[1]=false;
				//#*# Nous nettoyons les canaux attribus  ce fichier, s'ils sont valides
				if((0<(info->mLeft[1]+1))&&(info->mLeft[1]<gChCount)){
					for(int k=0;k<gFrPerBuf;k++){
						info->dbuf[k*gChCount+info->mLeft[1]]=0;
					}
				}
				if((0<(info->mRight[1]+1))&&(info->mRight[1]<gChCount)){
					for(int k=0;k<gFrPerBuf;k++){
						info->dbuf[k*gChCount+info->mRight[1]]=0;
					}
				}
			}
		}
		break; 

/*	case kInputLeft2:
		{//#*# Clear the entries in the buffer for the old channel
			for(int k=0;k<gFrPerBuf;k++){
				info->dbuf[k*gChCount+info->mLeft[1]]=0;
			}
			//#*# Set new left channel number
			info->mLeft[1] = (inNewValue->u.ivalue-1);
		}
		break;
	
	case kInputRight2:
		{//#*# Clear the entries in the buffer for the old channel
			for(int k=0;k<gFrPerBuf;k++){
				info->dbuf[k*gChCount+info->mRight[1]]=0;
			}
			//#*# Set new right channel number
			info->mRight[1] = (inNewValue->u.ivalue-1);
		}
		break;*/
		
	case kInputLoopEnable2:
		{//#*# Set new boolean value
			info->mLoopEnable[1] = (inNewValue->u.ivalue);
		}
		break;

	case kInputVolume2:
		{//#*# Set new volume value
			info->mVolume[1] = (inNewValue->u.fvalue);
		}
		break;

	case kInputStart2:
		{	if(!(info->isFileOpen[1])){
				if((info->sf[1] = sf_open(info->fileName[1], SFM_READ, &info->sfi[1])) == NULL){
					sf_close(info->sf[1]);
					info->isPlaying[1]=false;
				}else{
					info->isFileOpen[1]=true;
					info->isPlaying[1]=true;
				}
			}else{
				info->isPlaying[1]=true;
			}
		}
		break;

	case kInputSoundNum3:
		{//#*# Set new song number
			info->mSoundIndex[2] = (int)(inNewValue->u.ivalue);
			if(info->mPosition[2].u.fvalue!=0){//#*#on retrace la position  zro
				info->mPosition[2].u.fvalue=0; 
				SetOutputPropertyValue_(ip, inActorInfo, kOutputPosition3, &(info->mPosition[2]));
			}
			if(info->mSoundIndex[2]>0){//#*# !!! Ici il faudrait aussi ajouter comme maximum la 
				                       //quantit totale de fichiers audio dans la liste de la fenetre 
				                       //media d'ISadora. A l'heure qu'il est je ne sais pas comment recuperer
				                       //ce nombre, et passer un nombre plus grand que cleui l fait bouger 
				                       //la fonction DisposeActor, dans la ligne
				                       //Free2DCharArray(info->fileName,4) car filename a un format inattendu aprs
				                       //qu'on essaye d'y crir un nom de fichier qui n'existe pas.
				//#*# si on avait laisser un fichier ouvert  cet endroit, ne le ferme d'abord.
				if(info->isFileOpen[2]){
					sf_close(info->sf[2]);
					info->isFileOpen[2]=false;
				}
				info->fileName[2][255]='\0';
				ExecuteMediaCommand_(ip, inActorInfo, 0, kMediaGetName, kMediaTypeAudio, info->mSoundIndex[2], (long) info->fileName[2], sizeof(info->fileName[2]));

				//#*# Cette partie a t amelior par Mark par message priv sur le forum. Elle 
				//#*# permet de rcuperer la route du nom d'un fichier  partir de l'index
				//#*# On rcupre ici la route et le nom du fichier

				char fpBuf[1024];

				Boolean success = GetUTF8PathToCurrentDocument_(ip, inActorInfo->mThing, sizeof(fpBuf), fpBuf);
				if (success==TRUE && IsRelativePathName(info->fileName[2])) {
					//strncat(fpBuf, info->fileName[2], sizeof(fpBuf));
					sprintf(fpBuf, "%s%s",fpBuf, info->fileName[2]);
					//strncpy(info->fileName[2],fpBuf, sizeof(fpBuf));
					sprintf(info->fileName[2],"%s",fpBuf);
				}
				
				if((info->sf[2] = sf_open(info->fileName[2], SFM_READ, &info->sfi[2])) != NULL){
					info->isFileOpen[2]=true;
				}
				else{
					sf_close(info->sf[2]);
					info->isFileOpen[2]=false;
				}
			}else{
				info->isPlaying[2]=false;
			}
			if(!(info->isFileOpen[2])){//#*# si la reproduction d'aucun fichier n'arrive par la suite.
				info->isPlaying[2]=false;
				//#*# Nous nettoyons les canaux attribus  ce fichier, s'ils sont valides
				if((0<(info->mLeft[2]+1))&&(info->mLeft[2]<gChCount)){
					for(int k=0;k<gFrPerBuf;k++){
						info->dbuf[k*gChCount+info->mLeft[2]]=0;
					}
				}
				if((0<(info->mRight[2]+1))&&(info->mRight[2]<gChCount)){
					for(int k=0;k<gFrPerBuf;k++){
						info->dbuf[k*gChCount+info->mRight[2]]=0;
					}
				}
			}
		}
		break; 

/*	case kInputLeft3:
		{//#*# Clear the entries in the buffer for the old channel
			for(int k=0;k<gFrPerBuf;k++){
				info->dbuf[k*gChCount+info->mLeft[2]]=0;
			}
			//#*# Set new left channel number
			info->mLeft[2] = (inNewValue->u.ivalue-1);
		}
		break;
	
	case kInputRight3:
		{//#*# Clear the entries in the buffer for the old channel
			for(int k=0;k<gFrPerBuf;k++){
				info->dbuf[k*gChCount+info->mRight[2]]=0;
			}
			//#*# Set new right channel number
			info->mRight[2] = (inNewValue->u.ivalue-1);
		}
		break;*/
		
	case kInputLoopEnable3:
		{//#*# Set new boolean value
			info->mLoopEnable[2] = (inNewValue->u.ivalue);
		}
		break;

	case kInputVolume3:
		{//#*# Set new volume value
			info->mVolume[2] = (inNewValue->u.fvalue);
		}
		break;

	case kInputStart3:
		{	if(!(info->isFileOpen[2])){
				if((info->sf[2] = sf_open(info->fileName[2], SFM_READ, &info->sfi[2])) == NULL){
					sf_close(info->sf[2]);
					info->isPlaying[2]=false;
				}else{
					info->isFileOpen[2]=true;
					info->isPlaying[2]=true;
				}
			}else{
				info->isPlaying[2]=true;
			}
		}
		break;

	case kInputSoundNum4:
		{//#*# Set new song number
			info->mSoundIndex[3] = (int)(inNewValue->u.ivalue);
			if(info->mPosition[3].u.fvalue!=0){//#*#on retrace la position  zro
				info->mPosition[3].u.fvalue=0; 
				SetOutputPropertyValue_(ip, inActorInfo, kOutputPosition4, &(info->mPosition[3]));
			}
			if(info->mSoundIndex[3]>0){//#*# !!! Ici il faudrait aussi ajouter comme maximum la 
				                       //quantit totale de fichiers audio dans la liste de la fenetre 
				                       //media d'ISadora. A l'heure qu'il est je ne sais pas comment recuperer
				                       //ce nombre, et passer un nombre plus grand que cleui l fait bouger 
				                       //la fonction DisposeActor, dans la ligne
				                       //Free2DCharArray(info->fileName,4) car filename a un format inattendu aprs
				                       //qu'on essaye d'y crir un nom de fichier qui n'existe pas.
				//#*# si on avait laisser un fichier ouvert  cet endroit, ne le ferme d'abord.
				if(info->isFileOpen[3]){
					sf_close(info->sf[3]);
					info->isFileOpen[3]=false;
				}
				info->fileName[3][255]='\0';
				ExecuteMediaCommand_(ip, inActorInfo, 0, kMediaGetName, kMediaTypeAudio, info->mSoundIndex[3], (long) info->fileName[3], sizeof(info->fileName[3]));

				//#*# Cette partie a t amelior par Mark par message priv sur le forum. Elle 
				//#*# permet de rcuperer la route du nom d'un fichier  partir de l'index
				//#*# On rcupre ici la route et le nom du fichier

				char fpBuf[1024];
				Boolean success = GetUTF8PathToCurrentDocument_(ip, inActorInfo->mThing, sizeof(fpBuf), fpBuf);
				if (success==TRUE && IsRelativePathName(info->fileName[3])) {
					//strncat(fpBuf, info->fileName[3], sizeof(fpBuf));
					sprintf(fpBuf, "%s%s",fpBuf, info->fileName[3]);
					//strncpy(info->fileName[3],fpBuf, sizeof(fpBuf));
					sprintf(info->fileName[3],"%s",fpBuf);
				}
				if((info->sf[3] = sf_open(info->fileName[3], SFM_READ, &info->sfi[3])) != NULL){
					info->isFileOpen[3]=true;
				}
				else{
					sf_close(info->sf[3]);
					info->isFileOpen[3]=false;
				}
			}else{
				info->isPlaying[3]=false;
			}
			if(!(info->isFileOpen[3])){//#*# si la reproduction d'aucun fichier n'arrive par la suite.
				info->isPlaying[3]=false;
				//#*# Nous nettoyons les canaux attribus  ce fichier, s'ils sont valides
				if((0<(info->mLeft[3]+1))&&(info->mLeft[3]<gChCount)){
					for(int k=0;k<gFrPerBuf;k++){
						info->dbuf[k*gChCount+info->mLeft[3]]=0;
					}
				}
				if((0<(info->mRight[3]+1))&&(info->mRight[3]<gChCount)){
					for(int k=0;k<gFrPerBuf;k++){
						info->dbuf[k*gChCount+info->mRight[3]]=0;
					}
				}
			}
		}
		break; 

/*	case kInputLeft4:
		{//#*# Clear the entries in the buffer for the old channel
			for(int k=0;k<gFrPerBuf;k++){
				info->dbuf[k*gChCount+info->mLeft[3]]=0;
			}
			//#*# Set new left channel number
			info->mLeft[3] = (inNewValue->u.ivalue-1);
		}
		break;
	
	case kInputRight4:
		{//#*# Clear the entries in the buffer for the old channel
			for(int k=0;k<gFrPerBuf;k++){
				info->dbuf[k*gChCount+info->mRight[3]]=0;
			}
			//#*# Set new right channel number
			info->mRight[3] = (inNewValue->u.ivalue-1);
		}
		break;*/

	case kInputLoopEnable4:
		{//#*# Set new boolean value
			info->mLoopEnable[3] = (inNewValue->u.ivalue);
		}
		break;

	case kInputVolume4:
		{//#*# Set new volume value
			info->mVolume[3] = (inNewValue->u.fvalue);
		}
		break;

	case kInputStart4:
		{	if(!(info->isFileOpen[3])){
				if((info->sf[3] = sf_open(info->fileName[3], SFM_READ, &info->sfi[3])) == NULL){
					sf_close(info->sf[3]);
					info->isPlaying[3]=false;
				}else{
					info->isFileOpen[3]=true;
					info->isPlaying[3]=true;
				}
			}else{
				info->isPlaying[3]=true;
			}
		}
		break;
}
}


// ---------------------------------------------------------------------------------
//		 GetActorDefinedArea
// ---------------------------------------------------------------------------------
//	If the mGetActorDefinedAreaProc in the ActorInfo struct points to this function,
//	it indicates to Isadora that the object would like to draw either an icon or else
//	an graphic representation of its function.
//
//	### This function uses the 'PICT' 0 resource stored with the plugin to draw an icon.
//  You should replace this picture (located in the Plugin Resources.rsrc file) with
//  the icon for your actor.
// 
static ActorPictInfo	gPictInfo = { false, nil, nil, 0, 0 };

static Boolean
GetActorDefinedArea(
	IsadoraParameters*			ip,			
	ActorInfo*					inActorInfo,
	SInt16*						outTopAreaWidth,			// returns the width to reserve for the top Actor Defined Area
	SInt16*						outTopAreaMinHeight,		// returns the minimum height of the top area
	SInt16*						outBotAreaHeight,			// returns the width to reserve for the bottom Actor Defined Area
	SInt16*						outBotAreaMinWidth)			// returns the minimum width of the bottom area
{
	if (!gPictInfo.mInitialized) {
		PrepareActorDefinedAreaPict_(ip, inActorInfo, 0, &gPictInfo);
	}
	
	// place picture in top area
	*outTopAreaWidth = gPictInfo.mWidth;
	*outTopAreaMinHeight = gPictInfo.mHeight;
	
	// don't draw anything in bottom area
	*outBotAreaHeight = 0;
	*outBotAreaMinWidth = 0;
	
	return true;
}

// ---------------------------------------------------------------------------------
//		 DrawActorDefinedArea
// ---------------------------------------------------------------------------------
//	If GetActorDefinedArea is defined, then this function will be called whenever
//	your ActorDefinedArea needs to be drawn.
//
//	Beacuse we are using the PICT 0 resource stored with this plugin, we can use
//	the DrawActorDefinedAreaPict_ supplied by the Isadora callbacks.
//
//  DrawActorDefinedAreaPict_ is Alpha Channel aware, so you can have nice
//	shading if you like.

static void
DrawActorDefinedArea(
	IsadoraParameters*			ip,
	ActorInfo*					inActorInfo,
	void*						/* inDrawingContext */,		// unused at present
	ActorDefinedAreaPart		inActorDefinedAreaPart,		// the part of the actor that needs to be drawn
	ActorAreaDrawFlagsT			/* inAreaDrawFlags */,		// actor draw flags
	Rect*						inADAArea,					// rect enclosing the entire Actor Defined Area
	Rect*						/* inUpdateArea */,			// subset of inADAArea that needs updating
	Boolean						inSelected)					// true if actor is currently selected
{
	if (inActorDefinedAreaPart == kActorDefinedAreaTop && gPictInfo.mInitialized) {
		DrawActorDefinedAreaPict_(ip, inActorInfo, inSelected, inADAArea, &gPictInfo);
	}
}
	
// ---------------------------------------------------------------------------------
//		 GetActorInfo
// ---------------------------------------------------------------------------------
//	This is function is called by to get the actor's class and ID, and to get
//	pointers to the all of the plugin functions declared locally.
//
//	All members of the ActorInfo struct pointed to by outActorParams have been
//	set to 0 on entry. You only need set functions defined by your plugin
//	
EXPORT_ void
GetActorInfo(
	void*				/* inParam */,
	ActorInfo*			outActorParams)
{
	// REQUIRED information
	outActorParams->mActorName							= kActorName;
	outActorParams->mClass								= kActorClass;
	outActorParams->mID									= kActorID;
	outActorParams->mCompatibleWithVersion				= kCurrentIsadoraCallbackVersion;
	
	// REQUIRED functions
	outActorParams->mGetActorParameterStringProc		= GetParameterString;
	outActorParams->mGetActorHelpStringProc				= GetHelpString;
	outActorParams->mCreateActorProc					= CreateActor;
	outActorParams->mDisposeActorProc					= DisposeActor;
	outActorParams->mActivateActorProc					= ActivateActor;
	outActorParams->mHandlePropertyChangeValueProc		= HandlePropertyChangeValue;
	
	// OPTIONAL FUNCTIONS
	outActorParams->mHandlePropertyChangeTypeProc		= NULL;
	outActorParams->mHandlePropertyConnectProc			= NULL;
	outActorParams->mPropertyValueToStringProc			= NULL;
	outActorParams->mPropertyStringToValueProc			= NULL;
	outActorParams->mGetActorDefinedAreaProc			= GetActorDefinedArea;
	outActorParams->mDrawActorDefinedAreaProc			= DrawActorDefinedArea;
	outActorParams->mMouseTrackInActorDefinedAreaProc	= NULL;
}

// ---------------------------------------------------------------------------------
//		 ProcessAudioBuffer
// ---------------------------------------------------------------------------------
//	#*# This is the code that does the actual processing of Audio Buffer. Modify
// this code to create your own filter.
//

static void
ProcessAudioBuffer(
	IsadoraParameters*	/* ip */,	// not used in this function, but needed to call PluginAssert_
	PluginInfo*			info,
	ImageBufferPtr		srcBuf,
	ImageBufferPtr		outBuf)
{
	UInt32* srcData = static_cast<UInt32*>(srcBuf->mBaseAddress);
	UInt32 srcStride = srcBuf->mRowBytes - srcBuf->mWidth * sizeof(UInt32);

	UInt32* outData = static_cast<UInt32*>(outBuf->mBaseAddress);
	UInt32 outStride = outBuf->mRowBytes - outBuf->mWidth * sizeof(UInt32);
}

// ---------------------------------------------------------------------------------
//		 ReceiveMessage
// ---------------------------------------------------------------------------------
//	Isadora broadcasts messages to its Message Receives depending on what message
//	they are listening to. In this case, we are listening for kWantVideoFrameTick,
//	which is broadcast periodically (30 times per second.) When we receive the
//	message, we check to see if our video frame needs to be updated. If so, we
//	process the incoming video and pass the newly generated frame to the output.

static void
ReceiveMessage(
	IsadoraParameters*	ip,
	MessageMask			/* inMessageMask */,		// the message that caused this ReceiveMessage to be called -- one of the kWant... constants
	PortIndex			/* inIndex1 */,				// for MIDI messages, the port from which the message arrived. n/a otherwise
	const MsgData*		/* inData */,				// the data associated with this message
	UInt32				/* inLen */,				// the length of the data associated with this message
	long				inRefCon)					// in our use, actually the pointer to our ActorInfo
{
	// Convert the refCon into the ActorInfo* that it
	// really is, so that we can get at our data
	ActorInfo* actorInfo = reinterpret_cast<ActorInfo*>(inRefCon);

	// get pointer to plugin info
	PluginInfo* info = GetPluginInfo_(actorInfo);
	
	if(info->isDeviceRecognized){ //#*# si les MOTUs sont reconnues, on reproduit.
		bool b=false; //#*#variable de contrle pour voir si il y a de la reproduction sur au moins un canal
		for(int i=0;i<4;i++){
			if(info->isPlaying[i]){
				b=true;
				if(info->isFileOpen[i]){
					Value v={kBoolean,nil};
					v.u.ivalue=false;
					info->nread[i] = sf_readf_float(info->sf[i], info->buf[i], gFrPerBuf);//#*# read the audio file with sndfile
					if(info->nread[i]>0){
						for(int k=0;k<info->nread[i];k++){
							//#*#On crit l'audio aux canaux demands dans un buffer intermdiaire
							//#*#On ajuste le volume en pourcentage indiqu par l'utilisateur
							info->dbuf[k*gChCount+info->mLeft[i]]=(info->mVolume[i])*(info->buf[i][2*k])/100;
							info->dbuf[k*gChCount+info->mRight[i]]=(info->mVolume[i])*(info->buf[i][2*k+1])/100;
						}
						info->mPosition[i].u.fvalue=(info->mPosition[i].u.fvalue)+100*(info->nread[i])/(float)(info->sfi[i].frames);//#*#mis  jour de la position
						for(int k=info->nread[i];k<gFrPerBuf;k++){//#*#On crit le son nul l o on n'aurait pas crit de vrai son
							info->dbuf[k*gChCount+info->mLeft[i]]=0;
							info->dbuf[k*gChCount+info->mRight[i]]=0;
						}
					}
					else{//#*# Quand on arrive  la fin du fichier, 
						//on arrte de reproduire et on ferme le fichier
						//On le rouvre si l'option loop enable ait t choisie
						if(info->mLoopEnable[i]){
							sf_close(info->sf[i]);				
							if((info->sf[i] = sf_open(info->fileName[i], SFM_READ, &info->sfi[i])) == NULL){
								sf_close(info->sf[i]);
								info->isFileOpen[i]=false;
								info->isPlaying[i]=false;
							}
						}else{
							info->isPlaying[i]=false;						
							sf_close(info->sf[i]);
							info->isFileOpen[i]=false;
						}
						info->mPosition[i].u.fvalue=0;
						//#*#On signale par un trig que la fin du fichier a t atteinte
						v.u.ivalue=true;
					}
					switch(i){
						case(0):{
							SetOutputPropertyValue_(ip, actorInfo, kOutputPosition1, &(info->mPosition[0]));
							if(v.u.ivalue)SetOutputPropertyValue_(ip, actorInfo, kOutputEnd1, &v);
								}break;
						case(1):{
							SetOutputPropertyValue_(ip, actorInfo, kOutputPosition2, &(info->mPosition[1]));
							if(v.u.ivalue)SetOutputPropertyValue_(ip, actorInfo, kOutputEnd2, &v);
								}break;
						case(2):{
							SetOutputPropertyValue_(ip, actorInfo, kOutputPosition3, &(info->mPosition[2]));
							if(v.u.ivalue)SetOutputPropertyValue_(ip, actorInfo, kOutputEnd3, &v);
								}break;
						case(3):{
							SetOutputPropertyValue_(ip, actorInfo, kOutputPosition4, &(info->mPosition[3]));
							if(v.u.ivalue)SetOutputPropertyValue_(ip, actorInfo, kOutputEnd4, &v);
								}break;
					}
				}
			}
		}
		if(b){//#*# Si il y a quoi reproduir, on crit le buffer intermdiaire dans la sortie audio
			info->err=Pa_WriteStream(info->stream, info->dbuf, gFrPerBuf);
		}
	}
}
